"""Handy utility functions"""

import importlib.resources
import json
import platform
import logging
from pathlib import Path
import os.path
from tkinter import _tkinter
from typing import Any, Optional, Callable, Mapping
from unicodedata import combining, normalize

import regex as re

logger = logging.getLogger(__package__)

TraversablePath = importlib.resources.abc.Traversable | Path

# Flag so application code can detect if within a pytest run - only use if really needed
# See: https://pytest.org/en/7.4.x/example/simple.html#detect-if-running-from-within-a-pytest-run
called_from_test = False  # pylint: disable=invalid-name


#
# Functions to check which OS is being used
def is_mac() -> bool:
    """Return true if running on Mac"""
    return _is_system("Darwin")


def is_windows() -> bool:
    """Return true if running on Windows"""
    return _is_system("Windows")


def is_x11() -> bool:
    """Return true if running on Linux"""
    return _is_system("Linux")


# mypy: disable-error-code="attr-defined"
def _is_system(system: str) -> bool:
    """Return true if running on given system

    Args:
        system: Name of system to check against
    """
    try:
        return _is_system.system == system
    except AttributeError as exc:
        _is_system.system = platform.system()
        if _is_system.system not in ["Darwin", "Linux", "Windows"]:
            raise RuntimeError("Unknown windowing system") from exc
        return _is_system.system == system


def load_dict_from_json(filename: str) -> Optional[dict[str, Any]]:
    """If file exists, attempt to load into dict.

    Args:
        filename: Name of JSON file to load.

    Returns:
        Dictionary if loaded successfully, or None.
    """
    if os.path.isfile(filename):
        with open(filename, "r", encoding="utf-8") as fp:
            try:
                return json.load(fp)
            except json.decoder.JSONDecodeError as exc:
                logger.error(
                    f"Unable to load {filename} -- not valid JSON format\n" + str(exc)
                )
    return None


def load_wordfile_into_dict(path: TraversablePath, target_dict: dict) -> bool:
    """Load a one-word-per-line word list file into the target dictionary.

    Args:
        path: File to be loaded - accepts either a pathlib Path or
            an importlib.resources Traversable object.
        target_dict: Dictionary to be populated.

    Returns:
        True if file opened successfully, False if file not found.
    """
    try:
        try:
            with path.open("r", encoding="utf-8") as fp:
                data = fp.read()
        except UnicodeDecodeError:
            with path.open("r", encoding="iso-8859-1") as fp:
                data = fp.read()
        for line in data.split("\n"):
            word = line.strip()
            if word:
                target_dict[word] = True
        return True
    except FileNotFoundError:
        return False


class IndexRowCol:
    """Class to store/manipulate Tk Text indexes.

    Attributes:
        row: Row or line number.
        col: Column number.
    """

    row: int
    col: int

    def __init__(
        self, index_or_row: _tkinter.Tcl_Obj | str | int, col: Optional[int] = None
    ) -> None:
        """Construct index either from string or two ints.

        Args:
            index_or_row: Either int row, or index, which is string or Tcl_Obj that can be cast to string
            col: Optional column - needed if index_or_row is an int
        """
        if isinstance(index_or_row, _tkinter.Tcl_Obj):
            index_or_row = str(index_or_row)
        if isinstance(index_or_row, str):
            assert col is None
            rr, cc = index_or_row.split(".", 1)
            self.row = int(rr)
            self.col = int(cc)
        else:
            assert isinstance(index_or_row, int) and isinstance(col, int)
            self.row = index_or_row
            self.col = col

    def index(self) -> str:
        """Return string index from object's row/col attributes."""
        return f"{self.row}.{self.col}"

    def rowcol(self) -> tuple[int, int]:
        """Return row, col tuple."""
        return self.row, self.col

    def __eq__(self, other: object) -> bool:
        """Override equality test to check row and col."""
        if not isinstance(other, IndexRowCol):
            return NotImplemented
        return (self.row, self.col) == (other.row, other.col)


class IndexRange:
    """Class to store/manipulate a Text range defined by two indexes.

    Attributes:
        start: Start index.
        end: End index.
    """

    def __init__(
        self,
        start: _tkinter.Tcl_Obj | str | IndexRowCol,
        end: _tkinter.Tcl_Obj | str | IndexRowCol,
    ) -> None:
        """Initialize IndexRange with two Tcl_Obj/strings or IndexRowCol objects.


        Args:
            start: Index of start of range - either string or IndexRowCol/Tcl_Obj
            end: Index of end of range - either string or IndexRowCol/Tcl_Obj
        """
        if isinstance(start, _tkinter.Tcl_Obj):
            start = str(start)
        if isinstance(end, _tkinter.Tcl_Obj):
            end = str(end)
        if isinstance(start, str):
            self.start = IndexRowCol(start)
        else:
            assert isinstance(start, IndexRowCol)
            self.start = start
        if isinstance(end, str):
            self.end = IndexRowCol(end)
        else:
            assert isinstance(end, IndexRowCol)
            self.end = end

    def __eq__(self, other: object) -> bool:
        """Override equality test to check start and end of range."""
        if not isinstance(other, IndexRange):
            return NotImplemented
        return (self.start, self.end) == (other.start, other.end)


def sing_plur(count: int, singular: str, plural: str = "") -> str:
    """Return singular/plural phrase depending on count.

    Args:
        count: Number of items.
        singular: Singular version of item name.
        plural: Plural version of item name (default - add `s` to singular).

    Examples:
        sing_plur(1, "word") -> "1 word"
        sing_plur(2, "error") -> "2 errors"
        sing_plur(3, "match", "matches") -> "3 matches"
    """
    if count == 1:
        word = singular
    elif plural == "":
        word = singular + "s"
    else:
        word = plural
    return f"{count} {word}"


# Store callback that sounds bell, and provide function to call it.
# This is necessary since the bell requires various Tk features/widgets,
# like root and the status bar. We don't want to have to import those
# into every module that wants to sound the bell, e.g. Search.
_bell_callback = None  # pylint: disable=invalid-name


def bell_set_callback(callback: Callable[[], None]) -> None:
    """Register a callback function that will sound the bell.

    Args:
        callback: Bell-sounding function."""
    global _bell_callback
    _bell_callback = callback


def sound_bell() -> None:
    """Call the registered bell callback in order to sound the bell."""
    assert _bell_callback is not None
    _bell_callback()


def process_label(label: str) -> tuple[int, str]:
    """Convert a button label string, e.g. "~Save...", where the optional
    tilde indicates the underline location for keyboard activation,
    to the tilde location (-1 if none), and the string without the tilde.

    Args:
        label: Label to appear on widget, e.g. button.

    Returns:
        Tuple containing location of tilde in label string (-1 if none),
            and string with tilde removed.
    """
    return (label.find("~"), label.replace("~", ""))


def process_accel(accel: str) -> tuple[str, str]:
    """Convert accelerator string, e.g. "Ctrl+X" to appropriate keyevent
    string for platform, e.g. "Control-X".

    "Cmd/Ctrl" means use ``Cmd`` key on Mac; ``Ctrl`` key on Windows/Linux.

    Args:
        accel: Accelerator string.

    Returns:
        Tuple containing accelerator string and key event string suitable
        for current platform.
    """
    if is_mac():
        accel = accel.replace("/Ctrl", "")
    else:
        accel = accel.replace("Cmd/", "")
    keyevent = accel.replace("Ctrl+", "Control-")
    keyevent = keyevent.replace("Shift+", "Shift-")
    if is_mac():
        keyevent = keyevent.replace("Cmd+", "Command-")
        keyevent = keyevent.replace("Option+", "Option-")
        keyevent = keyevent.replace("Alt+", "Option-")
    else:
        keyevent = keyevent.replace("Alt+", "Alt-")
    accel = accel.replace("Key-", "")
    return (accel, f"<{keyevent}>")


def folder_dir_str(lowercase: bool = False) -> str:
    """Return "Folder" or "Directory" depending on platform.

    Args:
        lowercase: If True, return lowercase version.
    """
    fd_string = "Directory" if is_x11() else "Folder"
    return fd_string.lower() if lowercase else fd_string


def cmd_ctrl_string() -> str:
    """Return "Command" or "Control" depending on platform."""
    return "Command" if is_mac() else "Control"


class DiacriticRemover:
    """Supports removal of diacritics from strings."""

    outliers: Mapping
    source_outliers = "ä  æ  ǽ  đ ð ƒ ħ ı ł ø ǿ ö  œ  ß  ŧ ü  Ä  Æ  Ǽ  Đ Ð Ƒ Ħ I Ł Ø Ǿ Ö  Œ  ẞ  Ŧ Ü  Þ  þ"
    target_outliers = "ae ae ae d d f h i l o o oe oe ss t ue AE AE AE D D F H I L O O OE OE SS T UE TH th"

    @classmethod
    def setup_outliers(cls, source_outliers: str, target_outliers: str) -> None:
        """Setup the outliers mapping from source & target strings.

        Should be called with the relevant outliers when the language is changed."""
        DiacriticRemover.outliers = str.maketrans(
            dict(zip(source_outliers.split(), target_outliers.split()))
        )

    @classmethod
    def remove_diacritics(cls, string: str) -> str:
        """Remove accents, etc., from string.

        Based on https://stackoverflow.com/a/71408065
        First fixes a few outliers, like ǿ --> o, then uses Unicode Normalization
        to decompose string, then discards combining characters.
        """
        try:
            DiacriticRemover.outliers
        except AttributeError:
            DiacriticRemover.setup_outliers(
                DiacriticRemover.source_outliers, DiacriticRemover.target_outliers
            )
        return "".join(
            char
            for char in normalize("NFD", string.translate(DiacriticRemover.outliers))
            if not combining(char)
        )


class TextWrapper:
    """Provides Knuth-Plass rewrapping functionality.

    Can be replaced with Python's textwrap/TextWrapper if "greedy" wrapping is preferred:
        ``from textwrap import TextWrapper``

        ``wrapper = TextWrapper(break_long_words=False, break_on_hyphens=False)``
    """

    def __init__(self) -> None:
        """Initialize rewrap parameters."""

        # The four attributes below match the Perl version's four
        # options. However the Python version is normally passed
        # just three of them, namely:
        #
        #    width
        #    first_indent
        #    subsequent_indent
        #
        # The Python version will calculate a default value for
        # 'optimum_width' (see 'def fill' below) unless the caller
        # sets a value for this attribute.
        #
        # Only consider setting a value for 'optimum_width' if you
        # want to explore different optimal rewrappings. Using the
        # default value means that Knuth-Plass wrapping behaviour
        # is indentical in GG1 & GG2.

        self.width = 72
        self.optimum_width = -1
        self.initial_indent = ""
        self.subsequent_indent = ""

    def fill(self, paragraph: str) -> str:
        """Python implementation of ReflowGG.pm (Knuth-Plass)."""

        ####
        # The following are declared and accessed as 'nonlocal' variables
        # by functions defined within the enclosing scope of this top-most
        # function definition. The names must exist when first encountered
        # as they cannot be created by a first assignment in a nested def.
        ####

        # Maximum possible line length. Caller sets this to the current
        # 'rightmargin' value.
        maximum = self.width
        # Not normally passed by the caller so set default as per Perl
        # version, namely 'width - rmargindiff'. The latter is normally
        # '1'. Deal with situation when 'optimum_width' is set by the
        # caller to change the default value.
        if self.optimum_width == -1:
            # Caller hasn't changed it so set default value for optimum.
            optimum = [self.width - 1]
        else:
            # Use value set by caller for optimum.
            optimum = [self.optimum_width]
        # Indentation for first line.
        indent1 = self.initial_indent
        # Indentation for each line after the first.
        indent2 = self.subsequent_indent

        # Used by reflow_trial() in its calculation of the best breaks
        # in lines to produce the optimal rewrapping.
        penaltylimit = 0x2000000
        # Will hold the newline-terminated lines that are split from
        # the passed in paragraph text. This is the source of the
        # lines to be rewrapped by the Knuth algorithm.
        from_lines = []
        # Will hold the trimmed and newline-terminated lines that are
        # the result of the Knuth rewrapping.
        to_lines = []
        # Used by reflow_trial() to pick the break for the last line
        # which gives the least penalties for previous lines.
        lastbreak = 0
        # Is the list of breaks determined by reflow_trial. Will be used
        # by print_output() to construct the each of the rewrapped lines.
        linkbreak = []
        # The list of words from the passed in paragraph that are used
        # in the calculations by the Knuth algorithm to build rewrapped
        # lines with optimal breaks.
        words: list[str] = []
        # The count of words in the above list.
        wordcount = 0
        # Length of each word (excluding spaces) in the above list.
        word_len: list[int] = []
        # Length of the space after each word in word_len.
        space_len: list[int] = []
        # Used as an interim buffer of rewrapped lines that are edited
        # (e.g. to reinsert spaces in ellipses, etc.) before being
        # written to 'to_lines'.
        output: list[str] = []

        # NB The variables above are declared in this enclosing function's
        #    scope so that they exist when encountered as Python 'nonlocal'
        #    variables within the nested functions that follow.

        def reflow_trial() -> int:
            """Finds the optimal Knuth-Plass wrapping for a text
            within given margins and suggested optimum width.
            """

            nonlocal optimum, maximum, wordcount, penaltylimit
            nonlocal word_len, space_len, linkbreak

            totalpenalty = [0] * wordcount
            best = penaltylimit * 21
            best_lastbreak = 0
            for opt in optimum:
                my_linkbreak = [0] * wordcount
                j = 0
                # Optimize preceding break
                while j < wordcount:
                    interval = 0
                    totalpenalty[j] = penaltylimit * 2
                    k = j
                    while k >= 0:
                        interval += word_len[k]
                        if k < j and ((interval > opt + 10) or (interval >= maximum)):
                            break
                        penalty = (interval - opt) * (interval - opt)
                        interval += space_len[k]
                        if k > 0:
                            penalty += totalpenalty[k - 1]
                        if penalty < totalpenalty[j]:
                            totalpenalty[j] = penalty
                            my_linkbreak[j] = k - 1
                        k -= 1
                    j += 1
                interval = 0
                bestsofar = penaltylimit * 20
                lastbreak = wordcount - 2
                # Pick a break for the last line which gives the least
                # penalties for previous lines.
                k = wordcount - 2
                while k >= -1:
                    interval += word_len[k + 1]
                    if (interval > opt + 10) or (interval > maximum):
                        break
                    # Don't make last line too long.
                    if interval > opt:
                        penalty = (interval - opt) * (interval - opt)
                    else:
                        penalty = 0
                    interval += space_len[k + 1]
                    if k >= 0:
                        penalty += totalpenalty[k]
                    if penalty <= bestsofar:
                        bestsofar = penalty
                        lastbreak = k
                    k -= 1
                # Save these breaks if they are an improvement.
                if bestsofar < best:
                    best_lastbreak = lastbreak
                    linkbreak = my_linkbreak.copy()
                    best = bestsofar
                # Next opt
            return best_lastbreak

        def paragraph_to_lines(paragraph: str) -> None:
            """Split paragraph string into lines, keeping trailing empty lines.
            The split is on newlines and then they are restored, being careful
            not to add an extra newline at the end.

            Args:
                paragraph: String to be rewrapped. Usually one or more lines
                           each terminated with a newline.
            """

            nonlocal from_lines

            lines = paragraph.split("\n")
            # Remove any additional newline at the end.
            if len(lines) > 0 and lines[-1] == "":
                lines.pop()
            # Restore the terminating newlines.
            lines_with_newlines = map(lambda s: f"{s}\n", lines)
            # Create a global list of lines.
            from_lines = list(lines_with_newlines)

        def get_line() -> str:
            """Returns the next element of global array 'from_lines'. Assumes
            that no element in that list is zero length.
            """

            if len(from_lines) > 0:
                tmp = from_lines[0]
                del from_lines[0]
            else:
                tmp = ""
            return tmp

        def print_lines(lines: list[str]) -> None:
            """Trim EOL spaces and store the lines in the output buffer."""

            nonlocal to_lines

            for line in lines:
                to_lines.append(re.sub(r"[ \t]+\n", r"\n", line))

        def reflow_penalties() -> None:
            """Initialise global word- and space-length lists"""

            nonlocal wordcount, words, word_len, space_len

            wordcount = len(words)
            # Add paragraph indentation to first word if there is one.
            if wordcount > 0:
                words[0] = f"{indent1}{words[0]}"
            word_len = [0] * wordcount  # Length of each word (excluding spaces).
            space_len = [0] * wordcount  # Length of the space after after this word.
            for j in range(wordcount):
                if re.findall(r" $", words[j]):
                    word_len[j] = len(words[j]) - 1
                    space_len[j] = 2
                else:
                    word_len[j] = len(words[j])
                    space_len[j] = 1
            # First word already has 'indent1' added and will not be indented further.
            if wordcount > 0:
                word_len[0] -= len(indent2)

        def compute_output() -> None:
            """Compute global 'output' list from 'wordcount', 'words' list,
            lastbreak and linkbreak"""

            nonlocal output, words, wordcount, lastbreak, linkbreak

            output = []
            terminus = wordcount - 1
            while terminus >= 0:
                # NB The lower bound is inclusive, the upper bound is noninclusive.
                my_str = " ".join(words[lastbreak + 1 : terminus + 1]) + "\n"
                output.append(my_str)
                terminus = lastbreak
                lastbreak = linkbreak[lastbreak]
            output.reverse()
            # Add the indent to all but the first line.
            mapped = map(lambda s: f"{indent2}{s}", output[1 : len(output)])
            output = [output[0]]
            for m in mapped:
                output.append(m)

        def reflow_para() -> None:
            """Rewrap pragraph"""

            nonlocal words
            nonlocal lastbreak, linkbreak, wordcount, output

            if len(words) == 0:
                return
            reflow_penalties()
            lastbreak = 0
            lastbreak = reflow_trial()
            compute_output()
            # Restore spaces in ellipses
            tmp = []
            for line in output:
                tmp.append(re.sub(r"\x9f", " ", line))
            output = tmp.copy()
            print_lines(output)
            words = []

        def process(line: str) -> None:
            """Process a line by appending each word to global 'words' list.
            If the line is blank, reflow the paragraph of words in 'words'.
            """
            nonlocal words

            # Protect " . . ." ellipses by replacing space with unused byte \x9f.
            line = re.sub(r" \. \. \.", r"\x9f.\x9f.\x9f.", line)
            line = re.sub(r"\. \. \.", r".\x9f.\x9f.", line)
            # Splitting assumes a single space between words in 'line'
            line = re.sub(r"\s+", " ", line)
            # Remove any extra space at end of line.
            line = re.sub(r"\s$", "", line)
            linewords = line.split(" ")
            if len(linewords) != 0 and linewords[0] == "":
                del linewords[0]
            # If no words on this line then end of paragraph.
            if len(linewords) == 0:
                # No words on this line so end of paragraph.
                reflow_para()
                # Function expects a list of strings, each terminated with \n.
                print_lines([f"{indent1}\n"])
            else:
                # Add contents of global 'linewords' list to global 'words' list.
                for word in linewords:
                    words.append(word)

        ####
        # Executable section of enclosing function 'fill'.
        ####

        # Empty paragraphs don't need wrapping
        if paragraph == "":
            return ""

        # Split the passed in paragraph into lines and add them
        # to the global list 'from_lines'.
        paragraph_to_lines(paragraph)

        # Adjust global variables 'optimum' and 'maximum' by
        # 'indent2' length.
        if indent2 != "":
            maximum = maximum - len(indent2)
            mapped = map(lambda s: s - len(indent2), optimum)
            optimum = list(mapped)

        # Do the rewrapping.
        while line := get_line():
            process(line)
        # Reflow any remaining words in 'words' list.
        reflow_para()
        rewrapped_para = ""
        for line in to_lines:
            rewrapped_para = "".join((rewrapped_para, line))
        # If input paragraph did not have a terminating newline, output shouldn't
        if paragraph[-1] != "\n":
            rewrapped_para = rewrapped_para.rstrip("\n")
        return rewrapped_para



"""Illustration (Illo) or Sidenote (SN) checking, fixing and tidying functionality"""

import logging
import tkinter as tk
from tkinter import ttk
from typing import Optional, Any

import regex as re

from guiguts.checkers import CheckerDialog
from guiguts.maintext import maintext
from guiguts.misc_tools import tool_save
from guiguts.utilities import (
    IndexRowCol,
    IndexRange,
)

logger = logging.getLogger(__package__)

_the_illo_checker: Optional["IlloSNChecker"] = None  # pylint: disable=invalid-name
_the_sn_checker: Optional["IlloSNChecker"] = None  # pylint: disable=invalid-name


class IlloSNRecord:
    """Record of Illo or SN in file."""

    def __init__(
        self,
        text: str,
        start: IndexRowCol,
        end: IndexRowCol,
        mid_para: bool,
        hilite_start: int,
        hilite_end: int,
    ) -> None:
        """Initialize IlloSNRecord.

        Args:
            text - text of (first line of) the SN or Illo (if caption present).
            start - start rowcol of SN or Illo in file.
            end - end rowcol of SN or Illon in file.
            mid_para - True if Illo or SN record mid-paragraph, False otherwise.
            hilite_start - start column of highlighting in text.
            hilite_end - end column of highlighting in text.
        """
        self.text = text
        self.start = start
        self.end = end
        self.mid_para = mid_para
        self.hilite_start = hilite_start
        self.hilite_end = hilite_end


class SelectedIlloSNRecord:
    """Record of start & end line numbers (e.g. '5.0') for selected Illo or SN record"""

    def __init__(self, first: str, last: str) -> None:
        self.first_line_num = first
        self.last_line_num = last


class IlloSNCheckerDialog(CheckerDialog):
    """Illo or Sidenote Checker dialog."""

    def __init__(self, tag_type: str, **kwargs: Any) -> None:
        """Initialize Illo or Sidenote Checker dialog."""

        super().__init__(
            f"{tag_type} Check Results",
            tooltip="\n".join(
                [
                    f"Left click: Select & find {tag_type} tag",
                    "Right click: Remove item from list",
                    "Shift-Right click: Remove all matching items",
                ]
            ),
            **kwargs,
        )
        frame = ttk.Frame(self.header_frame)
        frame.grid(column=0, row=1, sticky="NSEW")
        self.move_up_btn = ttk.Button(
            frame,
            text="Move Selection Up",
            command=lambda: move_selection_up(tag_type),
        )
        self.move_up_btn.grid(column=0, row=0, sticky="NSW")
        self.move_down_btn = ttk.Button(
            frame,
            text="Move Selection Down",
            command=lambda: move_selection_down(tag_type),
        )
        self.move_down_btn.grid(column=1, row=0, sticky="NSW")


class IlloCheckerDialog(IlloSNCheckerDialog):
    """Illustration Fixup dialog."""

    manual_page = "Tools_Menu#Illustration_Fixup"

    def __init__(self, **kwargs: Any) -> None:
        """Initialize Illustration Fixup dialog."""
        super().__init__("Illustration", **kwargs)


class SNCheckerDialog(IlloSNCheckerDialog):
    """Sidenote Fixup dialog."""

    manual_page = "Tools_Menu#Sidenote_Fixup"

    def __init__(self, **kwargs: Any) -> None:
        """Initialize Sidenote Fixup dialog."""
        super().__init__("Sidenote", **kwargs)


class IlloSNChecker:
    """Find, check & record Illos or SNs."""

    def __init__(self, checker_dialog: IlloSNCheckerDialog) -> None:
        """Initialize IlloSNChecker."""
        self.illosn_records: list[IlloSNRecord] = []
        self.checker_dialog: IlloSNCheckerDialog = checker_dialog

    def reset(self) -> None:
        """Reset IlloSNChecker."""
        self.illosn_records = []

    def get_illosn_records(self) -> list[IlloSNRecord]:
        """Return the list of Illo or SN records."""
        return self.illosn_records

    def get_selected_illosn_record(
        self, selected_illosn_index: int
    ) -> SelectedIlloSNRecord:
        """Get line number indexes of the selected Illo or SN record.

        The returned indexes will be the same if selected record is not multi-line;
        e.g. [Illustration]

        Returns:
            Line number index of first line of the record; e.g. '5.0'
            Line number index of last line of the record; e.g. '6.14'
        """
        # We built a list of Illo or SN records with start/end RowCol details.
        records = self.get_illosn_records()
        # Get record of the selected tag
        selected_record = records[selected_illosn_index]
        # Get line number index of the first line of the selected record.
        selected_record_start = selected_record.start.index()
        first_line_num = maintext().index(f"{selected_record_start} linestart")
        # Get line number index of the last line of the selected record.
        # If not a multi-line record then this will be same index as above.
        selected_record_end = selected_record.end.index()
        last_line_num = maintext().index(f"{selected_record_end}")
        rec = SelectedIlloSNRecord(first_line_num, last_line_num)
        return rec

    def get_selected_illosn_index(self) -> int:
        """Get the index of the selected Illo or SN record.

        Returns:
            Index into self.illosn_records array, negative if none selected."""
        cur_idx = self.checker_dialog.current_entry_index()
        if cur_idx is None:
            return -1
        text_range = self.checker_dialog.entries[cur_idx].text_range
        assert text_range is not None
        illosn_start = text_range.start
        illosn_records = self.get_illosn_records()
        for illosn_index, illosn_record in enumerate(illosn_records):
            if illosn_record.start == illosn_start:
                return illosn_index
        return -1

    def check_for_anomalous_illo_or_sn(
        self, start: IndexRowCol, end: IndexRowCol
    ) -> bool:
        """Checks for hidden mid-paragraph Sidenotes, and other anomalies.

        The DP Formatting Guidelines allow a Project Manager to request
        that sidenotes are put next to the sentence they apply to in which
        case they are not separated out by blank lines nor prefixed with
        "*". Check for these 'hidden' mid-paragraph sidenotes.

        As a by-product it will also flag as 'MIDPARAGRAPH' other unusual
        formatting of Illo and SN tags in the file.

        Args:
            start - the start RowCol Index of the Illo or SN record.

        Returns:
            True if a 'hidden' mid-paragraph sidenote or other anomaly, False otherwise.
        """
        # An Illo or SN record is normally separated from other file lines by
        # a blank line above and below or one of those blank lines replaced
        # by a Page Marker record. Flag as mid-paragraph any other separation
        # of an Illo or SN record, such as by a blank line on one side and
        # a non-blank line on the other which isn't a Page Marker record.
        #
        file_last_line_num = maintext().end().row
        illosn_first_line_num = start.row
        illosn_last_line_num = end.row
        # Anomalous if line above illo/SN is not blank or page separator (except for first line of file)
        if illosn_first_line_num > 1:
            above_illosn_line_txt = maintext().get(
                f"{illosn_first_line_num - 1}.0", f"{illosn_first_line_num - 1}.end"
            )
            if not (
                above_illosn_line_txt == ""
                or above_illosn_line_txt.startswith("-----File:")
            ):
                return True

        # Find first line below illo/SN that is not blank/[Blank page]/page separator/illo/sidenote
        in_illo_sn = False
        for line_num in range(illosn_last_line_num + 1, file_last_line_num + 1):
            below_illosn_line_txt = maintext().get(f"{line_num}.0", f"{line_num}.end")
            if (
                below_illosn_line_txt == ""
                or below_illosn_line_txt == "[Blank Page]"
                or below_illosn_line_txt.startswith("-----File:")
            ):
                continue
            # Check if starting illo/SN markup. If on single line, flag is reset below
            if re.match(r"\*?\[(Illustration|Sidenote)", below_illosn_line_txt):
                in_illo_sn = True
            # If in illo/SN, keep on ignoring lines
            if in_illo_sn:
                # If illo/SN ends, reset the flag
                if "]" in below_illosn_line_txt:
                    in_illo_sn = False
                continue
            # Found "normal" text line - anomalous if previous line is not blank,
            # i.e. not a paragraph start, and ok if previous line is blank
            return maintext().get(f"{line_num - 1}.0", f"{line_num - 1}.end") != ""

        # Hit end of file before finding "normal" text, so not anomalous
        return False

    def run_check(self, tag_type: str) -> None:
        """Run the initial Illustration or Sidenote records check."""
        self.reset()
        search_range = maintext().start_to_end()
        match_regex = (
            r"\[ *illustration" if tag_type == "Illustration" else r"\[ *sidenote"
        )
        # Loop to find all tags of the type specified by argument 'tag_type'.
        # Allow some flexibility in spacing after the opening "[" of the tag.
        while begin_illosn_match := maintext().find_match(
            match_regex, search_range, regexp=True, nocase=True
        ):
            start = begin_illosn_match.rowcol
            # Find colon position, or use end of the required word (i.e. tag_type).
            colon_match = maintext().find_match(
                ":", IndexRange(start, maintext().rowcol(f"{start.row}.end"))
            )
            if colon_match is None:
                colon_pos = maintext().rowcol(
                    f"{start.index()}+{begin_illosn_match.count}c"
                )
            else:
                colon_pos = colon_match.rowcol
            end_point = colon_pos
            # Find closing square bracket - allow open/close bracket within an
            # Illustration or Sidenote record.
            end_match = None
            nested = False
            while True:
                end_match = maintext().find_match(
                    "[][]", IndexRange(end_point, maintext().end()), regexp=True
                )
                if end_match is None:
                    break
                end_point = maintext().rowcol(f"{end_match.rowcol.index()}+1c")
                if maintext().get_match_text(end_match) == "]":
                    if not nested:
                        break  # Found the one we want
                    nested = False  # Closing nesting
                else:
                    if nested:
                        end_match = None  # Not attempting to handle double nesting
                        break
                    nested = True  # Opening nesting

            # If closing [ not found, use end of line.
            if end_match is None:
                end_point = maintext().rowcol(f"{start.row}.end")
            # Get text of (first line of) the Illo or SN line(s) in the file. Note
            # that this does not include any prefixing '*' (that was added by the
            # formatting rounds to show that it is a mid-paragraph tag).
            illosn_txt = maintext().get(f"{start.index()}", f"{start.row}.end")
            # Look for a prefixing '*' to the tag.
            illosn_line = maintext().get(f"{start.index()}-1c", f"{start.row}.end")
            if illosn_line[0:1] == "*":
                mid_para = True
                illosn_txt = illosn_line
            else:
                # The tag will also be flagged as 'MIDPARAGRAPH' in the dialog if there
                # is any doubt that it is *not* a mid-paragraph tag.
                mid_para = self.check_for_anomalous_illo_or_sn(start, end_point)
            illosn_rec = IlloSNRecord(
                illosn_txt, start, end_point, mid_para, 1, colon_pos.col - start.col
            )
            self.illosn_records.append(illosn_rec)
            search_range = IndexRange(end_point, maintext().end())

    def update_after_move(self, tag_type: str, selected_illosn_index: int) -> None:
        """Update Illo or SN records list, update dialog and reselect tag just moved.

        Args:
            tag_type: either "Illustration" or "Sidenote"
        """
        # Update illosn_records list
        self.run_check(tag_type)
        # Update dialog
        display_illosn_entries(tag_type)
        # Select again the tag we have just moved so it is highlighted.
        self.checker_dialog.select_entry_by_index(selected_illosn_index)

    def delete_illosn_record_from_file(self, selected: SelectedIlloSNRecord) -> None:
        """Delete the selected Illo/SN record from its original location in the file.

        Do this from the start of (its first) line so that any prefixing '*' is
        included in the deletion. There are 3 cases to consider:
        1. The tag is immediately above a bottom of page Page Marker. Delete the
        tag line(s) and the preceding blank line.
        2. The tag is immediately below a bottom of page Page Marker. Delete the
        tag line(s) and the following blank line.
        3. The tag is mid-page so preceded and followed by a blank line. Delete
        the tag line(s) and the following blank line but leave the preceding
        one.
        """
        # Use marks to get updated index for start & end
        first_line_num = self.updated_index(selected.first_line_num)
        last_line_num = self.updated_index(selected.last_line_num)
        # Line before the (first record of the) Illo or SN record.
        prev_line_num = maintext().index(f"{first_line_num}-1l")
        prev_line_txt = maintext().get(prev_line_num, f"{prev_line_num} lineend")
        # Line after the (last record of the) Illo or SN record.
        next_line_num = maintext().index(f"{last_line_num}+1l linestart")
        next_line_txt = maintext().get(next_line_num, f"{next_line_num} lineend")
        # Assume it's a mid-page Illo or SN record unless found otherwise. This is case 3.
        if prev_line_txt != "":
            # Check this line. It's immediately above the (first line of the) Illo or SN record.
            if prev_line_txt[0:10] == "-----File:" and next_line_txt == "":
                # Case 2
                first_delete_line_num = maintext().index(f"{first_line_num}")
                last_delete_line_num = maintext().index(f"{last_line_num}+2l linestart")
                maintext().delete(first_delete_line_num, last_delete_line_num)
                return
        if next_line_txt != "":
            if next_line_txt[:10] == "-----File:" and prev_line_txt == "":
                # Case 1
                first_delete_line_num = maintext().index(f"{first_line_num}-1l")
                last_delete_line_num = maintext().index(f"{last_line_num}+1l linestart")
                maintext().delete(first_delete_line_num, last_delete_line_num)
                return
        if prev_line_txt == "" and next_line_txt == "":
            # Case 3
            first_delete_line_num = maintext().index(f"{first_line_num}")
            last_delete_line_num = maintext().index(f"{last_line_num}+2l linestart")
            maintext().delete(first_delete_line_num, last_delete_line_num)
            return
        # We get here if anomalous line spacing around the Illo or SN record.
        # Just delete the Illo or SN record and ignore any surrounding lines.
        maintext().delete(first_line_num, f"{last_line_num}+1l linestart")
        return

    def advance_up_to_first_bl_of_block(self, line_num: str) -> str:
        """Move insert point to the top of the block of blank lines.

        The insert point will not move if the 'block' is a single blank line.

        Arg:
            line_num: a line number index of a blank line; e.g. '5.0'

        Returns:
            A line number index, possibly same as the input one.
        """
        # Don't go past start of file.
        stop_line_num = "1.0"
        # We are at a blank line, possibly the bottom line of a block of
        # blank lines. Find the blank line at the top of the block and
        # return its line number.
        while True:
            if line_num == stop_line_num:
                # First line of file is a blank line (and top of the block).
                return line_num
            # Decrement line_num and see if it is another blank line.
            prev_line_num = line_num
            line_num = maintext().index(f"{line_num}-1l")
            if maintext().get(line_num, f"{line_num} lineend") == "":
                # We are in a block of blank lines. Continue looking for the
                # top of the block.
                continue
            # We get here if at the line above the top of the bl block.
            return prev_line_num

    def advance_down_to_last_bl_of_block(self, line_num: str) -> str:
        """Move insert point to the bottom of the block of blank lines.

        The insert point will not move if the 'block' is a single blank line.

        Arg:
            line_num: a line number index of a blank line; e.g. '5.0'

        Returns:
            A line number index, possibly same as the input one.
        """
        # Don't go past end of file.
        stop_line_num = maintext().end().index()
        # We are at a blank line, possibly the top line of a block of
        # blank lines. Find the blank line at the bottom of the block
        # and return its line number.
        while True:
            if line_num == stop_line_num:
                # Last line of file is a blank line (and bottom of the block).
                return line_num
            # Increment line_num and see if it is another blank line.
            prev_line_num = line_num
            line_num = maintext().index(f"{line_num}+1l")
            if maintext().get(line_num, f"{line_num} lineend") == "":
                # We are in a block of blank lines. Continue looking for the
                # bottom of the block.
                continue
            # We get here if at the line below the bottom of the bl block.
            return prev_line_num

    def updated_index(self, old_index: str) -> str:
        """Get an updated index using the checker dialog marks.

        If mark doesn't exist for, it's because of a "*" on the first line, so
        we just adjust the index here.

        Args:
            old_index: Old value for index.

        Returns:
            new value for index.
        """
        mark = self.checker_dialog.mark_from_rowcol(IndexRowCol(old_index))
        try:
            new_index = maintext().index(mark)
        except tk.TclError:
            new_index = maintext().index(re.sub(r"\.0$", ".1", mark))
            new_index = re.sub(r"\.1$", ".0", new_index)
        return new_index


def move_selection_up(tag_type: str) -> None:
    """Move selected Illo/SN tag up towards the start of the file.

    We won't move a tag past another tag of the same type.

    Args:
        tag_type: a string which is either "Sidenote" or "Illustration".
    """
    checker = _the_illo_checker if tag_type == "Illustration" else _the_sn_checker
    assert checker is not None
    selected_illosn_index = checker.get_selected_illosn_index()
    if selected_illosn_index < 0:
        return  # No selection.
    # Get the record for the selected tag from the records array we
    # built in run_check().
    selected = checker.get_selected_illosn_record(selected_illosn_index)
    # Group together all inserts/deletes that happen while moving the
    # selected tag.
    maintext().undo_block_begin()
    # Use marks to get updated index for start & end
    first_line_num = checker.updated_index(selected.first_line_num)
    last_line_num = checker.updated_index(selected.last_line_num)
    # Look for a suitable blank line. Start at line immediately above the (first line of
    # the) selected Illo or SN record. NB We may not be able to move the record at all if
    # it means skipping over another tag of the same type.
    line_num = maintext().index(f"{first_line_num}-1l")
    while (
        line_num != "0.0"
    ):  # NB That index does not exist but see at end of while loop.
        # What is on this line?
        line_txt = maintext().get(line_num, f"{line_num} lineend")
        # Is it a tag of the same type as the selected one?
        if tag_type == "Sidenote" and (
            line_txt[0:10] == "[Sidenote:" or line_txt[0:11] == "*[Sidenote:"
        ):
            # Can't move selected Sidenote above another Sidenote.
            return
        if tag_type == "Illustration" and (
            line_txt[0:14] == "[Illustration]"
            or line_txt[0:14] == "[Illustration:"
            or line_txt[0:15] == "*[Illustration]"
            or line_txt[0:15] == "*[Illustration:"
        ):
            # Can't move selected Illustration above another Illustration.
            return
        # No point in inserting the selected tag above the blank line in the situation below
        # as it will become the same place after the deletion of the Illo or SN record lines
        # from their original position.
        #
        # BL
        # ...
        # BL                     <- blank line we are at.
        # [Sidenote: Some text.] <- (first line of) the selected tag.
        # BL
        # ...
        #
        # If we are not in this situation, insert selected tag above the blank line we are at.
        # Otherwise look for next blank line above that one.
        if (
            line_txt == ""
            and not maintext().index(f"{line_num}+1l linestart") == first_line_num
        ):
            # If we are at the bottom of a block of blank lines, move the insertion
            # point to the top of the block. If the 'block' is just a single blank
            # line then the returned line_num is the same as the argument line_num.
            line_num = checker.advance_up_to_first_bl_of_block(line_num)
            # We can insert the selected tag above the blank line at line_num.
            # Copy the lines of the Illo or SN record to be moved. Don't include the
            # prefixing "*" if present.
            the_selected_record_lines = maintext().get(
                first_line_num, f"{last_line_num}"
            )
            if the_selected_record_lines[0:1] == "*":
                the_selected_record_lines = the_selected_record_lines[1:]
            # Delete the selected tag from its current position in the file.
            checker.delete_illosn_record_from_file(selected)
            # Insert copied lines above the blank line we are at and prefix it with a blank line.
            maintext().insert(
                maintext().index(f"{line_num}"),
                "\n" + the_selected_record_lines + "\n",
            )
            checker.update_after_move(tag_type, selected_illosn_index)
            return
        # If we have reached the top of the file, and it is not a blank line, then we
        # have to 'create' non-existent line number '0.0' to terminate the while-loop.
        if line_num == "1.0":
            line_num = "0.0"
        else:
            # Decrement line_num normally.
            line_num = maintext().index(f"{line_num}-1l")
    return


def move_selection_down(tag_type: str) -> None:
    """Move selected Illo/SN tag down towards the end of the file.

    We won't move a tag past another tag of the same type.

    Args:
        tag_type: a string which is either "Sidenote" or "Illustration".
    """
    checker = _the_illo_checker if tag_type == "Illustration" else _the_sn_checker
    assert checker is not None
    selected_illosn_index = checker.get_selected_illosn_index()
    if selected_illosn_index < 0:
        return  # No selection
    # Get the record for the selected tag from the records array we
    # built in run_check().
    selected = checker.get_selected_illosn_record(selected_illosn_index)
    # Group together all inserts/deletes that happen while moving the
    # selected tag.
    maintext().undo_block_begin()
    # Use marks to get updated index for start & end
    first_line_num = checker.updated_index(selected.first_line_num)
    last_line_num = checker.updated_index(selected.last_line_num)
    # Look for a suitable blank line. Start at line immediately below the (last line of
    # the) selected Illo or SN record. NB We may not be able to move the record at all if
    # it means skipping over another tag of the same type.
    line_num = maintext().index(f"{last_line_num}+1l linestart")
    # Set end of loop criteria.
    file_end = maintext().end().index()
    file_end_line_num_plus_1 = maintext().index(f"{file_end}+1l linestart")
    while line_num != file_end_line_num_plus_1:
        # What is on this line?
        line_txt = maintext().get(line_num, f"{line_num} lineend")
        # Is it a tag of the same type as the selected one?
        if tag_type == "Sidenote" and (
            line_txt[0:10] == "[Sidenote:" or line_txt[0:11] == "*[Sidenote:"
        ):
            # Can't move selected Sidenote above another Sidenote.
            return
        if tag_type == "Illustration" and (
            line_txt[0:14] == "[Illustration]"
            or line_txt[0:14] == "[Illustration:"
            or line_txt[0:15] == "*[Illustration]"
            or line_txt[0:15] == "*[Illustration:"
        ):
            # Can't move selected Illustration above another Illustration.
            return
        # No point in inserting the selected tag below the blank line in the situation below
        # as it will become the same place after the deletion of the Illo or SN record lines
        # from their original position.
        #
        # ...
        # BL
        # [Sidenote: Some text.] <- (last line of) the selected tag.
        # BL                     <- blank line we are at.
        # ...
        # BL
        #
        # If we are not in this situation, insert selected tag below the blank line we are at.
        # Otherwise look for next blank line below that one.
        if line_txt == "" and not maintext().index(
            f"{line_num}-1l linestart"
        ) == maintext().index(f"{last_line_num} linestart"):
            # If we are at the top of a block of blank lines, move the insertion
            # point to the bottom of the block. If the 'block' is just a single blank
            # line then the returned line_num is the same as the argument line_num.
            line_num = checker.advance_down_to_last_bl_of_block(line_num)
            # We can insert the selected tag below the blank line at line_num.
            # Copy the lines of the Illo or SN record to be moved. Don't include the
            # prefixing "*" if present.
            the_selected_record_lines = maintext().get(
                first_line_num, f"{last_line_num} lineend"
            )
            if the_selected_record_lines[0:1] == "*":
                the_selected_record_lines = the_selected_record_lines[1:]
            # Insert copied lines below the blank line we are at and suffix it with a blank line.
            maintext().insert(
                maintext().index(f"{line_num} lineend"),
                "\n" + the_selected_record_lines + "\n",
            )
            # Delete the selected tag from its current position in the file.
            checker.delete_illosn_record_from_file(selected)
            checker.update_after_move(tag_type, selected_illosn_index)
            return
        # Increment line_num.
        line_num = maintext().index(f"{line_num}+1l")
    return


def illosn_check(tag_type: str) -> None:
    """Check Illustration or Sidenote tags in the currently loaded file.

    Args:
        tag_type: which tag to check for - "Illustration" or "Sidenote"
    """
    global _the_illo_checker
    global _the_sn_checker

    if not tool_save():
        return
    assert tag_type in ("Illustration", "Sidenote")
    dialog_type = IlloCheckerDialog if tag_type == "Illustration" else SNCheckerDialog
    checker_dialog = dialog_type.show_dialog(
        rerun_command=lambda: illosn_check(tag_type),
        show_suspects_only=True,
        clear_on_undo_redo=True,
    )
    if tag_type == "Illustration":
        if _the_illo_checker is None:
            _the_illo_checker = IlloSNChecker(checker_dialog)
        elif not _the_illo_checker.checker_dialog.winfo_exists():
            _the_illo_checker.checker_dialog = checker_dialog
        the_checker = _the_illo_checker
    else:
        if _the_sn_checker is None:
            _the_sn_checker = IlloSNChecker(checker_dialog)
        elif not _the_sn_checker.checker_dialog.winfo_exists():
            _the_sn_checker.checker_dialog = checker_dialog
        the_checker = _the_sn_checker

    the_checker.run_check(tag_type)
    display_illosn_entries(tag_type)
    # Reselect correct entry in case this re-run was after an undo/redo
    checker_dialog.select_entry_after_undo_redo()
    # If that didn't cause anything to be selected, just select the first entry
    if checker_dialog.current_entry_index() is None:
        checker_dialog.select_entry_by_index(0)


def display_illosn_entries(tag_type: str) -> None:
    """(Re-)display the requested Illo/SN tag types in the checker dialog.

    Args:
        tag_type: which tag to check for - "Illustration" or "Sidenote"
    """
    assert tag_type in ("Illustration", "Sidenote")
    the_checker = _the_illo_checker if tag_type == "Illustration" else _the_sn_checker
    assert the_checker is not None
    checker_dialog = the_checker.checker_dialog
    checker_dialog.reset()
    illosn_records = the_checker.get_illosn_records()
    for illosn_record in illosn_records:
        error_prefix = ""
        if illosn_record.mid_para:
            error_prefix = "MIDPARAGRAPH: "
        checker_dialog.add_entry(
            illosn_record.text,
            IndexRange(illosn_record.start, illosn_record.end),
            illosn_record.hilite_start,
            illosn_record.hilite_end,
            error_prefix=error_prefix,
        )
    checker_dialog.display_entries(auto_select_line=False)



"""Jeebies check functionality"""

from enum import StrEnum, auto
from tkinter import ttk
from typing import List, Any

import importlib.resources
import logging
import regex as re

from guiguts.checkers import CheckerDialog, CheckerEntry
from guiguts.data import dictionaries
from guiguts.maintext import maintext
from guiguts.misc_tools import tool_save
from guiguts.utilities import cmd_ctrl_string
from guiguts.preferences import PersistentString, PrefKey, preferences
from guiguts.utilities import IndexRowCol, IndexRange

logger = logging.getLogger(__package__)

DEFAULT_DICTIONARY_DIR = importlib.resources.files(dictionaries)
DISCRIMINATOR_LEVEL = 1.0
HB_TOGGLE = {"h": "b", "H": "B", "b": "h", "B": "H"}


########################################################
# jeebies.py
# C author: Jim Tinsley (DP:jim) - 2005
# Go author: Roger Franks (DP:rfrank) - 2020
# Python author: Quentin Campbell (DP:qgc) - 2024
########################################################


class JeebiesParanoiaLevel(StrEnum):
    """Enum class to store Jeebies paranoia level types."""

    PARANOID = auto()
    NORMAL = auto()
    TOLERANT = auto()


class DictionaryNotFoundError(Exception):
    """Raised when no hebe-forms dictionary found."""

    def __init__(self, file: str) -> None:
        self.file = file


class JeebiesCheckerDialog(CheckerDialog):
    """Jeebies dialog."""

    manual_page = "Tools_Menu#Jeebies"

    def __init__(self, **kwargs: Any) -> None:
        """Initialize Jeebies dialog."""

        super().__init__(
            "Jeebies Results",
            tooltip="\n".join(
                [
                    "Left click: Select & find he/be error",
                    "Right click: Remove he/be error from list",
                    f"With {cmd_ctrl_string()} key: Also toggle queried he/be",
                    "Shift Right click: Also remove all matching he/be errors",
                ]
            ),
            **kwargs,
        )
        frame = ttk.Frame(self.header_frame)
        frame.grid(column=0, row=1, sticky="NSEW")
        ttk.Label(
            frame,
            text="Check Level:",
        ).grid(row=0, column=1, sticky="NSE", padx=(0, 5))
        self.paranoia_level = PersistentString(PrefKey.JEEBIES_PARANOIA_LEVEL)
        ttk.Radiobutton(
            frame,
            text="Paranoid",
            variable=self.paranoia_level,
            value=JeebiesParanoiaLevel.PARANOID,
            takefocus=False,
        ).grid(row=0, column=2, sticky="NSE", padx=2)
        ttk.Radiobutton(
            frame,
            text="Normal",
            variable=self.paranoia_level,
            value=JeebiesParanoiaLevel.NORMAL,
            takefocus=False,
        ).grid(row=0, column=3, sticky="NSE", padx=2)
        ttk.Radiobutton(
            frame,
            text="Tolerant",
            variable=self.paranoia_level,
            value=JeebiesParanoiaLevel.TOLERANT,
            takefocus=False,
        ).grid(row=0, column=4, sticky="NSE", padx=2)


class JeebiesChecker:
    """Provides jeebies check functionality."""

    def __init__(self) -> None:
        """Initialize SpellChecker class."""
        self.dictionary: dict[str, int] = {}
        self.load_phrases_file_into_dictionary()

    def check_for_jeebies_in_file(self) -> None:
        """Check for jeebies in the currently loaded file."""

        # Create the checker dialog to show results
        checker_dialog = JeebiesCheckerDialog.show_dialog(
            rerun_command=jeebies_check,
            process_command=self.process_jeebies,
        )

        # Check level used last time Jeebies was run or default if first run.
        check_level = preferences.get(PrefKey.JEEBIES_PARANOIA_LEVEL)

        # Get the paragraph strings and the ancillary lists that allow
        # us to map a hebe in a paragraph string to its actual line/col
        # position in the file.

        (
            input_lines,
            paragraph_strings,
            paragraph_start_line_numbers,
            paragraph_line_boundaries,
        ) = self.build_paragraph_structures()

        # Accumulate counts of 'be' and 'he' in the file.

        be_cnt_in_file = 0
        he_cnt_in_file = 0
        for line in input_lines:
            line_lc = line.lower()
            be_cnt_in_file += len(re.findall(r"(?<=\W|^)be(?=\W|$)", line_lc))
            he_cnt_in_file += len(re.findall(r"(?<=\W|^)he(?=\W|$)", line_lc))

        # We have a list of paragraphs as strings in paragraph_strings and counts
        # of 'he's and 'be's in the file.

        suspects_count = 0

        # Output header line to report. There will be two header lines if
        # the words 'he' and 'be' don't appear in a text. Pretty rare!

        if (be_cnt_in_file + he_cnt_in_file) == 0:
            checker_dialog.add_header(
                f"  --> 'be' counted {be_cnt_in_file} times and 'he' counted {he_cnt_in_file} times in file."
            )
            checker_dialog.add_header("")
            checker_dialog.add_header("    There are no he/be phrases to check.")
        else:
            checker_dialog.add_header(
                f"  --> 'be' counted {be_cnt_in_file} times and 'he' counted {he_cnt_in_file} times in file."
            )
            checker_dialog.add_header("")

            # For each paragraph in book ...

            for paragraph_number, paragraph_text in enumerate(paragraph_strings):
                # Search through paragraph text looking for three-word 'be' suspects.

                # Dialog functionality will rearrange dialog output so that line.col numbers
                # appear in correct sequence in the report after the six function calls below
                # are completed. Without that sorting it is possible that some line.col numbers
                # will be out of sequence. This occurs if one of the function calls reports a
                # suspect jeebie at the end of the paragraph then the next function call reports
                # a different suspect at the start of that paragraph. This can happen because
                # each function call rescans paragraph_text looking for a particular jeebies
                # suspect type.

                # There are two function calls below to identify and report suspect 3-word jeebie
                # phrases and four function calls to report suspect 2-word jeebie phrases; that is,
                # 'he' as first word of a 2-word phrase then 'he' as second word of a 2-word phrase
                # then repeated for 'be'.

                for he_be in ("he", "be"):
                    suspects_count += self.find_and_report_3_word_hebe_suspects(
                        he_be,
                        paragraph_text,
                        paragraph_number,
                        paragraph_start_line_numbers,
                        paragraph_line_boundaries,
                        input_lines,
                        checker_dialog,
                        check_level,
                    )
                    for first_second in ("first", "second"):
                        suspects_count += self.find_and_report_2_word_hebe_suspects(
                            he_be,
                            paragraph_text,
                            paragraph_number,
                            paragraph_start_line_numbers,
                            paragraph_line_boundaries,
                            input_lines,
                            checker_dialog,
                            first_second,
                            check_level,
                        )

            # Tell user if no suspect hebe phrases found in the paragraphs.

            if suspects_count == 0:
                checker_dialog.add_footer("    No suspect phrases found.")

        checker_dialog.display_entries()

    def build_paragraph_structures(self) -> tuple[List, List, List, List]:
        """Make the paragraph strings and the ancillary lists that allow
        us to map a hebe in a paragraph string to its actual line/col
        position in the file.

        """

        # Get the whole of the file from the main text widget
        input_lines = maintext().get_text().splitlines()
        # Ensure last paragraph converts to a line of text
        input_lines.append("")

        # List of paragraphs as single strings, case preserved.
        paragraph_strings = []
        # The line number of the first line in each paragraph.
        paragraph_start_line_numbers = []
        # Line boundaries in each paragraph. It is a list of lists.
        paragraph_line_boundaries = []
        # List of line boundaries for current paragraph. Used only within
        # the function and is refreshed for each new paragraph. Contents
        # will be appended to paragraph_line_boundaries list (see above).
        paragraph_line_ends: List[int] = []

        prev_line_was_blank = True
        line_number = 1
        s = ""
        line_end = 0
        open_paragraph = False

        for line in input_lines:
            # Is this input line blank?
            if re.match(r"^ *$", line):
                # It is blank.
                if open_paragraph:
                    # Close the paragraph we are constructing and save it.
                    # NB A blank line has been appended to input_lines to
                    #    ensure that the last paragraph being constructed
                    #    is always closed and saved.
                    paragraph_strings.append(s.rstrip())
                    # List of lists.
                    paragraph_line_boundaries.append(paragraph_line_ends)
                    paragraph_line_ends = []
                    s = ""
                    open_paragraph = False
                prev_line_was_blank = True
            elif prev_line_was_blank:
                # Not a blank line. Previous line was blank so start of a paragraph.
                # NB prev_line_was_blank is initially set True. This means that if
                #    very first line of the file is not blank then it is treated as
                #    the start of the first paragraph (i.e. line_number is 1).
                paragraph_start_line_numbers.append(line_number)
                open_paragraph = True
                # Start paragraph text string
                s = f"{line} "
                # Mark end of this first line in paragraph
                line_end = len(line) + 1
                paragraph_line_ends.append(line_end)
                prev_line_was_blank = False
            else:
                # Not a blank line and previous line was not blank so
                # continuation of a paragraph
                s = s + f"{line} "
                line_end = line_end + len(line) + 1
                paragraph_line_ends.append(line_end)

            line_number += 1

        return (
            input_lines,
            paragraph_strings,
            paragraph_start_line_numbers,
            paragraph_line_boundaries,
        )

    def find_and_report_2_word_hebe_suspects(
        self,
        hebe: str,
        paragraph_text: str,
        paragraph_number: int,
        paragraph_start_line_numbers: list[int],
        paragraph_line_boundaries: list[list[int]],
        file_lines_list: list[str],
        checker_dialog: JeebiesCheckerDialog,
        order: str,
        check_level: str,
    ) -> int:
        """Look for suspect "word be/he" or "be/he word" phrases in paragraphs."""

        def do_finditer_and_report(regx: str, para_lc: str) -> int:
            """Helper function for abstraction of repeated code."""

            count_of_suspects = 0
            for match_obj in re.finditer(regx, para_lc):
                if order == "first":
                    # Have a 2-word phrase starting with a hebe and prefixed by punctuation.
                    # E.g. '"Be careful!"' or ', he said.'
                    # See if it's in the list. If so get its frequency of occurrence.
                    hebe_form = para_lc[match_obj.start(0) : match_obj.end(0)]
                    hebe_count = self.find_in_dictionary(hebe_form)
                    hebe_start = match_obj.start(0) + hebe_form.index(hebe)
                    # Swap 'he' for 'be' (or vice versa) in 2-word phrase and lookup that.
                    # Avoid the "be behoven" or "he headed" trap.
                    s_hebe = f"{hebe} "
                    s_behe = f"{behe} "
                    behe_form = re.sub(s_hebe, s_behe, hebe_form)
                    behe_count = self.find_in_dictionary(behe_form)
                elif order == "second":
                    # Have a 2-word phrase ending with a hebe followed by punctuation.
                    # E.g. 'could be.'
                    # See if it's in the list. If so get its frequency of occurrence.
                    hebe_form = para_lc[match_obj.start(0) : match_obj.end(0)]
                    hebe_count = self.find_in_dictionary(hebe_form)
                    # Avoid the "the he-goats" trap when indexing for a hebe.
                    s_hebe = f" {hebe}"
                    # The result of 'hebe_form.index(s_hebe)' below will a pointer to the
                    # space in the 's_hebe' string, not the hebe itself. Add 1 so that
                    # 'hebe_start' correctly points at the hebe in the string.
                    hebe_start = match_obj.start(0) + hebe_form.index(s_hebe) + 1
                    # Swap 'he' for 'be' (or vice versa) in 2-word phrase and lookup that.
                    s_behe = f" {behe}"
                    behe_form = re.sub(s_hebe, s_behe, hebe_form)
                    behe_count = self.find_in_dictionary(behe_form)
                else:
                    # Shouldn't reach here.
                    hebe_count = behe_count = hebe_start = 0

                # At this point we have the two versions of the phrase. The variables
                # hebe_count and behe_count are integer frequencies of occurrence of
                # those phrases in a large corpus of DP texts.

                # The algorithm that follows improves on the Golang/PPWB method of identifying
                # suspect hebe phrases.

                if check_level == "tolerant" and (
                    hebe_count > 0 or hebe_count == 0 and behe_count == 0
                ):
                    # Even if the behe_count > hebe_count (see values calculation below) we
                    # won't query the phrase. That is, the 'tolerant' check passes if there
                    # are any 'good' occurrences or no 'bad' occurrences in the dictionary
                    # of examples.
                    continue

                if behe_count > 0 and (
                    hebe_count == 0 or behe_count / hebe_count > DISCRIMINATOR_LEVEL
                ):
                    count_of_suspects += 1
                    info = f"({behe_count}/{hebe_count})"

                    # Add this 2-form and its info to the dialog

                    # We have the start position in the paragraph string of this hebe.
                    # Now locate the actual file line and corresponding position on that
                    # line of this hebe.

                    line_number, hebe_position_on_line = self.locate_file_line(
                        paragraph_number,
                        hebe_start,
                        paragraph_start_line_numbers,
                        paragraph_line_boundaries,
                    )
                    line = file_lines_list[line_number - 1]

                    self.add_to_dialog(
                        info,
                        line,
                        line_number,
                        hebe_position_on_line,
                        checker_dialog,
                    )

                elif (
                    check_level in ("normal", "paranoid")
                    and hebe_count == 0
                    and behe_count == 0
                ):
                    # Neither 'he' nor 'be' versions of our 2-form phrase are in the
                    # dictionary.

                    count_of_suspects += 1
                    info = f"({behe_count}/{hebe_count})"

                    # Query this 2-form phrase in the report.

                    # We have the start position in the paragraph string of its hebe.
                    # Locate the corresponding file line and its column position on
                    # that line.

                    line_number, hebe_position_on_line = self.locate_file_line(
                        paragraph_number,
                        hebe_start,
                        paragraph_start_line_numbers,
                        paragraph_line_boundaries,
                    )
                    line = file_lines_list[line_number - 1]

                    self.add_to_dialog(
                        info,
                        line,
                        line_number,
                        hebe_position_on_line,
                        checker_dialog,
                    )

            return count_of_suspects

        ####
        # Find and report 2-word hebe suspects in paragraph text
        ####

        suspects_count = 0
        # Get the opposite word to the one being reported on.
        behe = "he" if hebe == "be" else "be"

        # Look for 2-form hebe phrases. There are two types:
        #   1. The hebe is the first word of the phrase: e.g. '“Be diligent ....”'
        #   2. The hebe is the second word of the phrase: e.g. '“... could be.”'
        # Note that the hebe in a 2-form phrase is ALWAYS delimited by at least
        # one punctuation character, as in the examples above, or a punctuation
        # character followed by a space as in '... had done. He then ...'. Special
        # treatment needed when phrase involves a contaction as in 'an’ be damned.'
        # See below for this.

        if order == "first":
            # Looking for hebe as first word of 2-word punctuation delimited phrase.

            para_lc = paragraph_text.lower()

            # NB A hebe phrase containing a contraction as in 'an’ be damned' has already
            #    been processed as a 3-form phrase. Avoid it being processed again as a
            #    2-form phrase. Obscure the right curly quote characters whenever they
            #    and a space precede a hebe.
            qs_hebe = f"’ {hebe}"
            ss_hebe = f"  {hebe}"
            para_lc = re.sub(qs_hebe, ss_hebe, para_lc)
            # Look for type 1, 2-form hebe phrases in the edited paragraph string.
            regx = rf"(?<=^|\p{{P}}|\p{{P}} ){hebe} [a-z]+"
            suspects_count += do_finditer_and_report(regx, para_lc)

        elif order == "second":
            # Looking for hebe as second word of 2-word punctuation delimited phrase.

            para_lc = paragraph_text.lower()

            regx = rf"[a-z]+ {hebe}(?=\p{{P}}|$)"
            suspects_count += do_finditer_and_report(regx, para_lc)

        else:
            # Should never get here. Variable 'order' value either 'first' or 'second'.
            pass

        return suspects_count

    def find_and_report_3_word_hebe_suspects(
        self,
        hebe: str,
        paragraph_text: str,
        paragraph_number: int,
        paragraph_start_line_numbers: list[int],
        paragraph_line_boundaries: list[list[int]],
        file_lines_list: list[str],
        checker_dialog: JeebiesCheckerDialog,
        check_level: str,
    ) -> int:
        """Look for suspect "w1 be w2" or "w1 he w2" phrases in paragraphs."""

        def make_dialog_line(info: str, checker_dialog: JeebiesCheckerDialog) -> None:
            """Helper function for abstraction of repeated code."""

            # We have the start position in the paragraph string of a suspect hebe.
            # Locate the corresponding file line and its column position on that
            # line.

            line_number, hebe_position_on_line = self.locate_file_line(
                paragraph_number,
                hebe_start,
                paragraph_start_line_numbers,
                paragraph_line_boundaries,
            )
            line = file_lines_list[line_number - 1]

            self.add_to_dialog(
                info, line, line_number, hebe_position_on_line, checker_dialog
            )

        ####
        # Find and report 3-word hebe suspects in paragraph text
        ####

        suspects_count = 0
        # Get the opposite word to the one being reported on.
        behe = "he" if hebe == "be" else "be"

        # Search for 3-form pattern "w1 he/be w2" in a lower-case copy of paragraph.
        para_lc = paragraph_text.lower()
        for match_obj in re.finditer(f"[a-z’]+ {hebe} [a-z’]+", para_lc):
            # Have a 3-word form here (e.g. "must be taken" or "long he remained").
            # See if it's in the list. If so get its frequency of occurrence.
            hebe_form = para_lc[match_obj.start(0) : match_obj.end(0)]
            hebe_count = self.find_in_dictionary(hebe_form)
            # Avoid the "tribe be not" or "catastrophe he had" trap.
            s_hebe = f" {hebe}"
            hebe_start = match_obj.start(0) + hebe_form.index(s_hebe) + 1
            # Swap 'he' for 'be' (or vice versa) in 3-word form and lookup that.
            # Avoid the "tribe be not" or "catastrophe he had" trap.
            s_behe = f" {behe}"
            behe_form = re.sub(s_hebe, s_behe, hebe_form)
            behe_count = self.find_in_dictionary(behe_form)

            # At this point we have the two versions of the phrase. The variables
            # hebe_count and behe_count are integer frequencies of occurrence of
            # those phrases in a large corpus of DP texts.

            # The algorithm that follows improves on the Golang/PPWB method of identifying
            # suspect hebe phrases.

            if check_level == "tolerant" and hebe_count > 0:
                # Even if the behe_count > hebe_count (see values calculation below) we
                # won't query the phrase.
                continue

            if behe_count > 0 and (
                hebe_count == 0 or behe_count / hebe_count > DISCRIMINATOR_LEVEL
            ):
                # Query this 3-form phrase.
                suspects_count += 1
                info = f"({behe_count}/{hebe_count})"

                # Query this 3-form phrase in the report.

                make_dialog_line(info, checker_dialog)

            elif check_level == "normal" and hebe_count == 0 and behe_count == 0:
                # Neither 'he' nor 'be' versions of our 3-form phrase are in the
                # dictionary. Does it contain a 2-form hebe phrase that is in
                # the dictionary? E.g. the 3-form phrase 'also be contracted' is
                # not in the dictionary but the 2-form hebe phrases 'also be' or
                # 'be contracted' could be. If either in the dictionary don't
                # query the 3-form phrase they're part of.

                # Isolate the 2-form phrases and see if either is in the dictionary.
                parts = hebe_form.split()
                phrase1 = parts[0] + " " + parts[1]
                phrase2 = parts[1] + " " + parts[2]
                if (
                    self.find_in_dictionary(phrase1) > 0
                    or self.find_in_dictionary(phrase2) > 0
                ):
                    continue

                # Neither 2-form phrase in the dictionary.
                suspects_count += 1
                info = f"({behe_count}/{hebe_count})"

                # Query this 3-form phrase in the report.

                make_dialog_line(info, checker_dialog)

            elif check_level == "paranoid" and hebe_count == 0 and behe_count == 0:
                # Neither 'he' nor 'be' versions of our 3-form phrase are in the
                # dictionary. Query it in report.
                suspects_count += 1
                info = f"({behe_count}/{hebe_count})"

                # Query this 3-form phrase in the report.

                make_dialog_line(info, checker_dialog)

        return suspects_count

    def locate_file_line(
        self,
        rec_num: int,
        hebe_start: int,
        paragraph_start_line_numbers: list[int],
        paragraph_line_boundaries: list[list[int]],
    ) -> tuple[int, int]:
        """Returns line number in file and position on that line of a hebe found in paragraph string."""
        paragraph_start_line_number = paragraph_start_line_numbers[rec_num]
        hebe_start_position_on_line = 0
        line_boundaries = paragraph_line_boundaries[rec_num]
        for indx, boundary in enumerate(line_boundaries):
            if hebe_start < boundary:
                # Located file line number
                line_number = paragraph_start_line_number + indx
                if indx == 0:
                    # We are on first line of paragraph
                    hebe_start_position_on_line = hebe_start
                else:
                    hebe_start_position_on_line = hebe_start - line_boundaries[indx - 1]
                break
        return line_number, hebe_start_position_on_line

    def find_in_dictionary(self, phrase: str) -> int:
        """Returns frequency in hebe phrase corpus or zero."""

        freq = self.dictionary.get(phrase, 0)
        if freq != 0:
            return freq
        # Try again with a prefixing " " character
        key = f" {phrase}"
        freq = self.dictionary.get(key, 0)
        if freq != 0:
            return freq
        # Try finally with a trailing " " character
        key = f"{phrase} "
        freq = self.dictionary.get(key, 0)
        return freq

    def add_to_dialog(
        self,
        info: str,  # pylint: disable=unused-argument
        line: str,
        line_number: int,
        hebe_start: int,
        checker_dialog: JeebiesCheckerDialog,
    ) -> None:
        """Helper function that abstracts repeated code.

        Args:
            info: Ratio of 'he' or 'be freq / 'be' or 'he' freq expressed as a string.
            line: Text of file line.
            line_num: Its line number in the file.
            hebe_start: The position on line of string 'he' or 'be'.
            checker_dialog: Where report text is written.
        """
        # Get start/end of 'he' or 'be' string in file.
        error_start = f"{line_number}.{hebe_start}"
        error_end = f"{line_number}.{hebe_start + 2}"
        # Store in structure for file row/col positions & ranges.
        start_rowcol = IndexRowCol(error_start)
        end_rowcol = IndexRowCol(error_end)

        ############################################################
        # To RESTORE ratios in report lines uncomment the following
        # two lines and comment out the third line.
        ############################################################
        # line = f"{info} {line}"
        # highlight_start = hebe_start + len(info) + 1
        highlight_start = hebe_start
        highlight_end = highlight_start + 2

        # Add record to the dialog.
        checker_dialog.add_entry(
            line,
            IndexRange(start_rowcol, end_rowcol),
            highlight_start,
            highlight_end,
        )

    def load_phrases_file_into_dictionary(self) -> None:
        """Load a phrase-per-line file into the target dictionary. Entries in the file
        are either 2-form (e.g. `|he|a:168`) or 3-form (e.g. `since|he|stopped:7`) phrases.
        """
        file = "hebelist.txt"
        path = DEFAULT_DICTIONARY_DIR.joinpath(file)
        try:
            with path.open("r", encoding="utf-8") as fp:
                for line in fp:
                    line = line.strip()
                    if ":" in line:
                        key, value = line.split(":")
                        key = key.replace("|", " ")
                        self.dictionary[key] = int(value)
        except FileNotFoundError as exc:
            raise DictionaryNotFoundError(file) from exc

    def process_jeebies(self, checker_entry: CheckerEntry) -> None:
        """Process the Jeebies query."""
        if checker_entry.text_range is None:
            return
        start_mark = JeebiesCheckerDialog.mark_from_rowcol(
            checker_entry.text_range.start
        )
        end_mark = JeebiesCheckerDialog.mark_from_rowcol(checker_entry.text_range.end)
        match_text = maintext().get(start_mark, end_mark)
        # Toggle match text
        replacement_text = HB_TOGGLE[match_text[0]] + match_text[1]
        maintext().replace(start_mark, end_mark, replacement_text)


def jeebies_check() -> None:
    """Check for jeebies in the currently loaded file."""

    if not tool_save():
        return
    try:
        JeebiesChecker().check_for_jeebies_in_file()
    except DictionaryNotFoundError as exc:
        logger.error(f"Dictionary not found: {exc.file}")
        return
